import javax.jdo.*;
import javax.jdo.annotations.PersistenceCapable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryUsage;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;
import static java.lang.Thread.sleep;
// javac -cp objectdb.jar Test.java
// java -cp objectdb.jar com.objectdb.Enhancer -cp . Test Test$Data
// java -cp objectdb.jar;. Test
public class Test{
    @PersistenceCapable private static class Data{
        @SuppressWarnings("unused") private int value;
        void setValue(int value){
            this.value = value;
        }
    }
    private static PersistenceManagerFactory persistenceManagerFactory;
    private static PersistenceManager persistenceManager;
    public static void main(String[] args){
        System.out.format("==========%nRunning test ...%n");
        printMemoryUsage();
        pause("Open this application in JavaVisualVm");
        openDatabase();
        printMemoryUsage();
        createDatabaseObjects();
        printMemoryUsage();
        pause("Inspect Heap Dump in JavaVisualVm.  Should show many of our Data objects.%nDo a full GC in " +
            "JavaVisualVm and all softly referenced objects, i.e. our Data objects, should be flushed from memory.");
        int choice = choose("Choose a flushing attempt technique:%n  0 - none%n  1 - Evict Objects%n  2 - " +
            "Delete Objects%n  3 - Close Factory");
        switch(choice){
            default:
            case 0:
                System.out.format("SKIP%n");
                break;
            case 1:
                System.out.format("EVICT%n");
                evictDatabaseObjects();
                break;
            case 2:
                System.out.format("DELETE%n");
                deleteDatabaseObjects();
                break;
            case 3:
                System.out.format("CLOSE%n");
                closeDatabase();
                break;
        }
        printMemoryUsage();
        pause("Inspect JavaVisualVm.  Data objects should have been flushed.");
        closeDatabase();
        printMemoryUsage();
        System.out.format("Test finished.%n==========%n");
    }
    private static void openDatabase(){
        System.out.format("openDatabase()%n");
        Path currentDirectoryPath = Paths.get(System.getProperty("user.dir", "."));
        Path objectDbConfPath = currentDirectoryPath.resolve("ObjectDb.conf");
        Path databasePath = currentDirectoryPath.resolve("Test.tmp");
        System.setProperty("objectdb.home", currentDirectoryPath.toString());
        System.setProperty("objectdb.conf", objectDbConfPath.toString());
        Properties properties = new Properties();
        properties.setProperty("javax.jdo.PersistenceManagerFactoryClass", "com.objectdb.jdo.PMF");
        properties.setProperty("javax.jdo.option.Optimistic", "true");
        properties.setProperty("javax.jdo.option.RetainValues", "true");
        properties.setProperty("javax.jdo.option.RestoreValues", "false");
        properties.setProperty("javax.jdo.option.IgnoreCache", "true");
        properties.setProperty("javax.jdo.option.NontransactionalRead", "false");
        properties.setProperty("javax.jdo.option.NontransactionalWrite", "false");
        properties.setProperty("javax.jdo.option.Multithreaded", "true");
        properties.setProperty("javax.jdo.option.DetachAllOnCommit", "false");
        properties.setProperty("javax.jdo.option.CopyOnAttach", "false");
        properties.setProperty("javax.jdo.option.TransactionType", "RESOURCE_LOCAL");
        properties.setProperty("javax.jdo.option.TransactionIsolationLevel", "read-uncommitted");
        properties.setProperty("javax.jdo.option.ConnectionURL", databasePath.toString() + ";drop");
        persistenceManagerFactory = JDOHelper.getPersistenceManagerFactory(properties);
        persistenceManager = persistenceManagerFactory.getPersistenceManager();
    }
    private static void closeDatabase(){
        System.out.format("closeDatabase()%n");
        while(!persistenceManagerFactory.isClosed()){
            try{
                persistenceManagerFactory.close();
            }catch(JDOUserException e){
                try{
                    sleep(100);
                }catch(InterruptedException ignore){}
            }
        }
    }
    private static void createDatabaseObjects(){
        System.out.format("createDatabaseObjects()%n");
        Transaction transaction = persistenceManager.currentTransaction();
        try{
            transaction.begin();
            for(int index = 0; index < 1_000; index++){
                Data data = new Data();
                data.setValue(index + 1);// +1 to match ObjectDB Explorer datastore ID
                persistenceManager.makePersistent(data);
            }
            transaction.commit();
        }catch(Exception e){
            e.printStackTrace();
        }finally{
            if(transaction.isActive()) transaction.rollback();
        }
    }
    private static void deleteDatabaseObjects(){
        System.out.format("deleteDatabaseObjects()%n");
        Transaction transaction = persistenceManager.currentTransaction();
        try{
            transaction.begin();
            Extent<Data> dataExtent = persistenceManager.getExtent(Data.class);
            for(Data data : dataExtent) persistenceManager.deletePersistent(data);
            transaction.commit();
        }catch(Exception e){
            e.printStackTrace();
        }finally{
            if(transaction.isActive()) transaction.rollback();
        }
    }
    private static void evictDatabaseObjects(){
        System.out.format("evictDatabaseObjects()%n");
        Transaction transaction = persistenceManager.currentTransaction();
        try{
            transaction.begin();
            Extent<Data> dataExtent = persistenceManager.getExtent(Data.class);
            for(Data data : dataExtent) persistenceManager.evict(data);
            transaction.commit();
        }catch(Exception e){
            e.printStackTrace();
        }finally{
            if(transaction.isActive()) transaction.rollback();
        }
    }
    private static void printMemoryUsage(){
        MemoryMXBean memoryMXBean = ManagementFactory.getMemoryMXBean();
        MemoryUsage memoryUsage = memoryMXBean.getHeapMemoryUsage();
        long heapUsed = memoryUsage.getUsed();// In bytes
        System.out.format("Application now using this much heap RAM:  %,d (%,d MB)%n", heapUsed, heapUsed / 1024 / 1024);
    }
    private static void pause(String message){
        System.out.format(message + "%n(Press ENTER to continue.)");
        try{
            //noinspection ResultOfMethodCallIgnored
            System.in.read();
        }catch(IOException e){
            e.printStackTrace();
        }
    }
    private static int choose(String message){
        int choice = 0;
        System.out.format(message + "%n(Press a number then ENTER to continue.)%n");
        try{
            BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(System.in));
            String line = bufferedReader.readLine();
            choice = Integer.parseInt(line);
        }catch(IOException e){
            e.printStackTrace();
        }
        return choice;
    }
}
